@@ -18,7 +18,9 @@ class DMG_MBC:
 	RAM_BANK_SIZE = 0x2000
 	ROM_BANK_NUM = 0
 	CURRENT_ROM_BANK = 0
+	CURRENT_FLASH_BANK = 0
 	START_BANK = 0
+	RTC_BUFFER = None
 
 	def __init__(self, args=None, cart_write_fncptr=None, cart_read_fncptr=None, cart_powercycle_fncptr=None, clk_toggle_fncptr=None):
 		if args is None: args = {}
@@ -37,11 +39,11 @@ class DMG_MBC:
 		mbc_id = args["mbc"]
 		if mbc_id in (0x01, 0x02, 0x03):						# 0x01:'MBC1', 0x02:'MBC1+SRAM', 0x03:'MBC1+SRAM+BATTERY',
 			return DMG_MBC1(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
-		elif mbc_id == 0x06:									# 0x06:'MBC2+SRAM+BATTERY',
+		elif mbc_id in (0x05, 0x06):							# 0x06:'MBC2+SRAM+BATTERY',
 			return DMG_MBC2(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
-		elif mbc_id in (0x10, 0x13):							# 0x10:'MBC3+RTC+SRAM+BATTERY', 0x13:'MBC3+SRAM+BATTERY',
+		elif mbc_id in (0x0F, 0x10, 0x11, 0x12, 0x13, 0x110):	# 0x10:'MBC3+RTC+SRAM+BATTERY', 0x13:'MBC3+SRAM+BATTERY',
 			return DMG_MBC3(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
-		elif mbc_id in (0x19, 0x1B, 0x1C, 0x1E):				# 0x19:'MBC5', 0x1B:'MBC5+SRAM+BATTERY', 0x1C:'MBC5+RUMBLE', 0x1E:'MBC5+RUMBLE+SRAM+BATTERY',
+		elif mbc_id in (0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E):	# 0x19:'MBC5', 0x1B:'MBC5+SRAM+BATTERY', 0x1C:'MBC5+RUMBLE', 0x1E:'MBC5+RUMBLE+SRAM+BATTERY',
 			return DMG_MBC5(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
 		elif mbc_id == 0x20:									# 0x20:'MBC6+FLASH+SRAM+BATTERY',
 			return DMG_MBC6(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
@@ -51,7 +53,7 @@ class DMG_MBC:
 			return DMG_MBC1M(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
 		elif mbc_id in (0x0B, 0x0D):							# 0x0B:'MMM01',  0x0D:'MMM01+SRAM+BATTERY',
 			return DMG_MMM01(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
-		elif mbc_id == 0xFC:									# 0xFC:'GBD+SRAM+BATTERY',
+		elif mbc_id == 0xFC:									# 0xFC:'MAC-GBD+SRAM+BATTERY',
 			return DMG_GBD(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
 		elif mbc_id == 0x105:									# 0x105:'G-MMC1+SRAM+BATTERY',
 			return DMG_GMMC1(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
@@ -73,6 +75,8 @@ class DMG_MBC:
 			return DMG_Unlicensed_Sachen(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
 		elif mbc_id == 0x205:									# 0x205:'Datel Orbit V2',
 			return DMG_Unlicensed_DatelOrbitV2(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
+		# elif mbc_id == 0x206:									# 0x206:'Datel MegaMem',
+		# 	return DMG_Unlicensed_DatelMegaMem(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
 		else:
 			self.__init__(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=clk_toggle_fncptr)
 			return self
@@ -105,6 +109,9 @@ class DMG_MBC:
 	def GetROMBank(self):
 		return self.CURRENT_ROM_BANK
 
+	def GetFlashBank(self):
+		return self.CURRENT_FLASH_BANK
+
 	def GetROMBanks(self, rom_size):
 		return math.ceil(rom_size / self.ROM_BANK_SIZE)
 	
@@ -120,6 +127,9 @@ class DMG_MBC:
 	def GetROMSize(self):
 		return self.ROM_BANK_SIZE * self.ROM_BANK_NUM
 	
+	def GetMaxROMSize(self):
+		return 32*1024
+
 	def CalcChecksum(self, buffer):
 		chk = 0
 		for i in range(0, len(buffer), 2):
@@ -141,9 +151,11 @@ class DMG_MBC:
 	def SelectBankROM(self, index):
 		dprint(self.GetName(), "|", index)
 		commands = [
-			[ 0x2100, index & 0xFF ]
+			[ 0x2100, index & 0xFF ],
 		]
+		
 		start_address = 0 if index == 0 else 0x4000
+
 		self.CartWrite(commands)
 		return (start_address, self.ROM_BANK_SIZE)
 	
@@ -159,6 +171,12 @@ class DMG_MBC:
 	def SetStartBank(self, index):
 		self.START_BANK = index
 	
+	def SelectBankFlash(self, index):
+		return
+	
+	def HasFlashBanks(self):
+		return False
+	
 	def HasHiddenSector(self):
 		return False
 	
@@ -177,6 +195,9 @@ class DMG_MBC:
 	def WriteRTC(self, buffer, advance=False):
 		pass
 
+	def GetRTCString(self):
+		return "Not available"
+
 	def ResetBeforeBankChange(self, index):
 		return False
 
@@ -216,6 +237,9 @@ class DMG_MBC1(DMG_MBC):
 		self.CartWrite(commands)
 		return (start_address, self.ROM_BANK_SIZE)
 
+	def GetMaxROMSize(self):
+		return 2*1024*1024
+
 class DMG_MBC2(DMG_MBC):
 	def GetName(self):
 		return "MBC2"
@@ -223,6 +247,9 @@ class DMG_MBC2(DMG_MBC):
 	def SelectBankRAM(self, index):
 		return (0, self.RAM_BANK_SIZE)
 
+	def GetMaxROMSize(self):
+		return 256*1024
+
 class DMG_MBC3(DMG_MBC):
 	def GetName(self):
 		return "MBC3"
@@ -236,9 +263,6 @@ class DMG_MBC3(DMG_MBC):
 	
 	def HasRTC(self):
 		dprint("Checking for RTC")
-		if self.MBC_ID != 16:
-			dprint("No RTC because wrong MBC ID", self.MBC_ID)
-			return False
 		self.EnableRAM(enable=False)
 		self.EnableRAM(enable=True)
 		self.LatchRTC()
@@ -248,13 +272,14 @@ class DMG_MBC3(DMG_MBC):
 			self.CLK_TOGGLE_FNCPTR(60)
 			self.CartWrite([ [0x4000, i] ])
 			data = self.CartRead(0xA000, 0x800)
+			if len(data) == 0: return False
 			if data[0] in (0, 0xFF): continue
 			skipped = False
 			if data != bytearray([data[0]] * 0x800):
 				dprint("No RTC because whole bank is not the same value", data[0])
 				skipped = True
 				break
-
+		
 		self.EnableRAM(enable=False)
 		self.CartWrite([ [0x4000, 0] ])
 		return skipped is False
@@ -290,6 +315,7 @@ class DMG_MBC3(DMG_MBC):
 		
 		self.EnableRAM(enable=False)
 		self.CartWrite([ [0x4000, 0] ])
+		self.RTC_BUFFER = buffer
 		return buffer
 
 	def WriteRTC(self, buffer, advance=False):
@@ -359,7 +385,7 @@ class DMG_MBC3(DMG_MBC):
 		self.CLK_TOGGLE_FNCPTR(50)
 		self.CartWrite([ [ 0x4000, 0x0C ] ])
 		self.CLK_TOGGLE_FNCPTR(50)
-		self.CartWrite([ [ 0xA000, 0x40 ] ])
+		self.CartWrite([ [ 0xA000, 0x40 ] ], sram=True)
 		time.sleep(0.1)
 		
 		# Write to registers
@@ -368,7 +394,7 @@ class DMG_MBC3(DMG_MBC):
 			self.CartWrite([ [ 0x4000, i ] ])
 			self.CLK_TOGGLE_FNCPTR(50)
 			data = struct.unpack("<I", buffer[(i-8)*4:(i-8)*4+4])[0] & 0xFF
-			self.CartWrite([ [ 0xA000, data ] ])
+			self.CartWrite([ [ 0xA000, data ] ], sram=True)
 		time.sleep(0.1)
 
 		# Latch RTC
@@ -380,6 +406,36 @@ class DMG_MBC3(DMG_MBC):
 
 		self.CartWrite([ [0x4000, 0] ])
 		self.EnableRAM(enable=False)
+	
+	def GetRTCString(self):
+		s = ""
+		if self.RTC_BUFFER is None:
+			self.ReadRTC()
+		rtc_buffer = self.RTC_BUFFER
+
+		rtc_s = rtc_buffer[0x00]
+		rtc_m = rtc_buffer[0x04]
+		rtc_h = rtc_buffer[0x08]
+		rtc_d = (rtc_buffer[0x0C] | rtc_buffer[0x10] << 8) & 0x1FF
+		rtc_carry = ((rtc_buffer[0x10] & 0x80) != 0)
+		if rtc_carry: rtc_d += 512
+		if rtc_h > 24 or rtc_m > 60 or rtc_s > 60:
+			try:
+				dprint("Invalid RTC state: {:d} days, {:02d}:{:02d}:{:02d}".format(rtc_d, rtc_h, rtc_m, rtc_s))
+			except:
+				pass
+			s = "Invalid state"
+		elif rtc_h == 0 and rtc_m == 0 and rtc_s == 0 and rtc_d == 0 and rtc_carry == 0:
+			s = "Not available"
+		else:
+			if rtc_d == 1:
+				s = "{:d} day, {:02d}:{:02d}:{:02d}".format(rtc_d, rtc_h, rtc_m, rtc_s)
+			else:
+				s = "{:d} days, {:02d}:{:02d}:{:02d}".format(rtc_d, rtc_h, rtc_m, rtc_s)
+		return s
+
+	def GetMaxROMSize(self):
+		return 4*1024*1024
 
 class DMG_MBC5(DMG_MBC):
 	def GetName(self):
@@ -401,21 +457,19 @@ class DMG_MBC5(DMG_MBC):
 	
 	def SelectBankROM(self, index):
 		dprint(self.GetName(), "|", index)
-		if index == 0 or index >= 256:
-			commands = [
-				[ 0x2100, index & 0xFF ],
-				[ 0x3000, ((index >> 8) & 0xFF) ],
-			]
-		else:
-			commands = [
-				[ 0x2100, index & 0xFF ],
-			]
+		commands = [
+			[ 0x3000, ((index >> 8) & 0xFF) ],
+			[ 0x2100, index & 0xFF ],
+		]
 		
 		start_address = 0 if index == 0 else 0x4000
 
 		self.CartWrite(commands)
 		return (start_address, self.ROM_BANK_SIZE)
 
+	def GetMaxROMSize(self):
+		return 8*1024*1024
+
 class DMG_MBC6(DMG_MBC):
 	def __init__(self, args=None, cart_write_fncptr=None, cart_read_fncptr=None, cart_powercycle_fncptr=None, clk_toggle_fncptr=None):
 		if args is None: args = {}
@@ -441,6 +495,9 @@ class DMG_MBC6(DMG_MBC):
 		start_address = 0 if index == 0 else 0x4000
 		return (start_address, self.ROM_BANK_SIZE)
 
+	def HasFlashBanks(self):
+		return True
+
 	def SelectBankFlash(self, index):
 		dprint(self.GetName(), "|", index)
 		self.CURRENT_ROM_BANK = index
@@ -504,7 +561,7 @@ class DMG_MBC6(DMG_MBC):
 			sr = self.CartRead(0x4000)
 			dprint("Status Register Check: 0x{:X} == 0x80? {:s}".format(sr, str(sr == 0x80)))
 			if sr == 0x80: break
-			time.sleep(0.0001)
+			time.sleep(0.01)
 
 	def GetFlashID(self):
 		self.EnableFlash(enable=True)
@@ -524,6 +581,9 @@ class DMG_MBC6(DMG_MBC):
 		self.SelectBankROM(self.CURRENT_ROM_BANK)
 		return flash_id
 
+	def GetMaxROMSize(self):
+		return 1*1024*1024
+
 class DMG_MBC7(DMG_MBC):
 	def GetName(self):
 		return "MBC7"
@@ -539,6 +599,9 @@ class DMG_MBC7(DMG_MBC):
 		]
 		self.CartWrite(commands)
 
+	def GetMaxROMSize(self):
+		return 4*1024*1024
+
 class DMG_MBC1M(DMG_MBC1):
 	def GetName(self):
 		return "MBC1M"
@@ -555,6 +618,9 @@ class DMG_MBC1M(DMG_MBC1):
 		self.CartWrite(commands)
 		return (start_address, self.ROM_BANK_SIZE)
 
+	def GetMaxROMSize(self):
+		return 1*1024*1024
+
 class DMG_MMM01(DMG_MBC):
 	def GetName(self):
 		return "MMM01"
@@ -596,9 +662,26 @@ class DMG_MMM01(DMG_MBC):
 		self.CartWrite(commands)
 		return (start_address, self.ROM_BANK_SIZE)
 
+	def GetMaxROMSize(self):
+		return 1*1024*1024
+
 class DMG_GBD(DMG_MBC5):
 	def GetName(self):
-		return "GBD"
+		return "MAC-GBD"
+
+	def SelectBankROM(self, index):
+		dprint(self.GetName(), "|", index)
+		commands = [
+			[ 0x2100, index & 0xFF ],
+		]
+		
+		start_address = 0 if index == 0 else 0x4000
+
+		self.CartWrite(commands)
+		return (start_address, self.ROM_BANK_SIZE)
+	
+	def GetMaxROMSize(self):
+		return 1*1024*1024
 
 class DMG_GMMC1(DMG_MBC5):
 	def GetName(self):
@@ -639,15 +722,10 @@ class DMG_GMMC1(DMG_MBC5):
 	
 	def SelectBankROM(self, index):
 		dprint(self.GetName(), "|", index)
-		if index == 0 or index >= 256:
-			commands = [
-				[ 0x2100, index & 0xFF ],
-			]
-		else:
-			commands = [
-				[ 0x2100, index & 0xFF ],
-			]
-		
+		commands = [
+			[ 0x2100, index & 0xFF ],
+		]
+
 		start_address = 0 if index == 0 else 0x4000
 
 		self.CartWrite(commands)
@@ -735,6 +813,9 @@ class DMG_GMMC1(DMG_MBC5):
 		else:
 			return super().CalcChecksum(buffer=buffer)
 
+	def GetMaxROMSize(self):
+		return 1*1024*1024
+
 class DMG_M161(DMG_MBC):
 	def GetName(self):
 		return "M161"
@@ -747,9 +828,6 @@ class DMG_M161(DMG_MBC):
 	def ResetBeforeBankChange(self, index):
 		return True
 	
-	#def GetROMSize(self):
-	#	return self.ROM_BANK_SIZE * math.floor(self.ROM_BANK_NUM / 2)
-
 	def SelectBankROM(self, index):
 		dprint(self.GetName(), "|", index)
 		commands = [
@@ -758,10 +836,24 @@ class DMG_M161(DMG_MBC):
 		self.CartWrite(commands)
 		return (0, 0x8000)
 
+	def GetMaxROMSize(self):
+		return 256*1024
+
 class DMG_HuC1(DMG_MBC5):
 	def GetName(self):
 		return "HuC-1"
 
+	def SelectBankROM(self, index):
+		dprint(self.GetName(), "|", index)
+		commands = [
+			[ 0x2100, index & 0xFF ],
+		]
+		
+		start_address = 0 if index == 0 else 0x4000
+
+		self.CartWrite(commands)
+		return (start_address, self.ROM_BANK_SIZE)
+	
 	def EnableRAM(self, enable=True):
 		dprint(self.GetName(), "|", enable)
 		commands = [
@@ -769,6 +861,9 @@ class DMG_HuC1(DMG_MBC5):
 		]
 		self.CartWrite(commands)
 
+	def GetMaxROMSize(self):
+		return 1*1024*1024
+
 class DMG_HuC3(DMG_MBC):
 	def GetName(self):
 		return "HuC-3"
@@ -786,6 +881,10 @@ class DMG_HuC3(DMG_MBC):
 			[ 0xA000, 0x60 ],
 			[ 0x0000, 0x0D ],
 			[ 0xA000, 0xFE ],
+		]
+		self.CartWrite(commands)
+		time.sleep(0.01)
+		commands = [
 			[ 0x0000, 0x0C ],
 			[ 0x0000, 0x00 ],
 
@@ -793,10 +892,14 @@ class DMG_HuC3(DMG_MBC):
 			[ 0xA000, 0x40 ],
 			[ 0x0000, 0x0D ],
 			[ 0xA000, 0xFE ],
+		]
+		self.CartWrite(commands)
+		time.sleep(0.01)
+		commands = [
 			[ 0x0000, 0x0C ],
 			[ 0x0000, 0x00 ]
 		]
-		self.CartWrite(commands, delay=0.01)
+		self.CartWrite(commands)
 
 		rtc = 0
 		for i in range(0, 6):
@@ -805,9 +908,13 @@ class DMG_HuC3(DMG_MBC):
 				[ 0xA000, 0x10 ],
 				[ 0x0000, 0x0D ],
 				[ 0xA000, 0xFE ],
+			]
+			self.CartWrite(commands)
+			time.sleep(0.01)
+			commands = [
 				[ 0x0000, 0x0C ]
 			]
-			self.CartWrite(commands, delay=0.01)
+			self.CartWrite(commands)
 			rtc |= (self.CartRead(0xA000) & 0x0F) << (i * 4)
 			self.CartWrite([[ 0x0000, 0x00 ]], delay=0.01)
 		
@@ -819,14 +926,15 @@ class DMG_HuC3(DMG_MBC):
 		
 		#dstr = ' '.join(format(x, '02X') for x in buffer)
 		#dprint("[{:02X}] {:s}".format(int(len(dstr)/3) + 1, dstr))
-
+		
+		self.RTC_BUFFER = buffer
 		return buffer
 
 	def WriteRTC(self, buffer, advance=False):
 		if advance:
 			try:
 				dt_now = datetime.datetime.fromtimestamp(time.time())
-				print(buffer)
+				dprint(buffer)
 				if buffer == bytearray([0x00] * len(buffer)): # Reset
 					hours = 0
 					minutes = 0
@@ -916,13 +1024,28 @@ class DMG_HuC3(DMG_MBC):
 		dstr = ' '.join(format(x, '02X') for x in buffer)
 		dprint("[{:02X}] {:s}".format(int(len(dstr)/3) + 1, dstr))
 	
+	def GetRTCString(self):
+		if self.RTC_BUFFER is None:
+			self.ReadRTC()
+		rtc_buffer = struct.unpack("<I", self.RTC_BUFFER[0:4])[0]
+		rtc_h = math.floor((rtc_buffer & 0xFFF) / 60)
+		rtc_m = (rtc_buffer & 0xFFF) % 60
+		rtc_d = (rtc_buffer >> 12) & 0xFFF
+		if rtc_d == 1:
+			return "{:d} day, {:02d}:{:02d}".format(rtc_d, rtc_h, rtc_m)
+		else:
+			return "{:d} days, {:02d}:{:02d}".format(rtc_d, rtc_h, rtc_m)
+
+	def GetMaxROMSize(self):
+		return 2*1024*1024
+	
 class DMG_TAMA5(DMG_MBC):
 	def GetName(self):
 		return "TAMA5"
 
 	def EnableMapper(self):
 		tama5_check = self.CartRead(0xA000)
-		lives = 150
+		lives = 20
 		while (tama5_check & 3) != 1:
 			dprint("- Current value is 0x{:X}, now writing 0xA001=0x{:X}".format(tama5_check, 0x0A))
 			self.CartWrite([[0xA001, 0x0A]], sram=True)
@@ -930,7 +1053,7 @@ class DMG_TAMA5(DMG_MBC):
 			time.sleep(0.01)
 			lives -= 1
 			if lives < 0:
-				dprint("Failed!")
+				print("Error: Couldn’t enable TAMA5 mapper!")
 				return False
 		dprint("Enabled TAMA5 successfully")
 		return True
@@ -998,6 +1121,7 @@ class DMG_TAMA5(DMG_MBC):
 		self.CartWrite(commands, sram=True)
 		self.SelectBankROM(0)
 
+		self.RTC_BUFFER = buffer
 		return buffer
 
 	def WriteRTC(self, buffer, advance=False):
@@ -1132,34 +1256,65 @@ class DMG_TAMA5(DMG_MBC):
 					value2 = value1
 					value1 = self.CartRead(0xA000)
 
+	def GetRTCString(self):
+		if self.RTC_BUFFER is None:
+			self.ReadRTC()
+		rtc_buffer = self.RTC_BUFFER
+		seconds = Util.DecodeBCD(rtc_buffer[0x00])
+		minutes = Util.DecodeBCD(rtc_buffer[0x01])
+		hours = Util.DecodeBCD(rtc_buffer[0x02])
+		#weekday = rtc_buffer[0x03] & 0xF
+		days = Util.DecodeBCD(rtc_buffer[0x03] >> 4 | (rtc_buffer[0x04] & 0xF) << 4)
+		months = Util.DecodeBCD(rtc_buffer[0x04] >> 4 | (rtc_buffer[0x05] & 0xF) << 4)
+		years = Util.DecodeBCD(rtc_buffer[0x05] >> 4 | (rtc_buffer[0x06] & 0xF) << 4)
+		leap_year_state = Util.DecodeBCD(rtc_buffer[0x0D] >> 4)
+		if leap_year_state == 0:
+			return "Year {:d}*, {:d}-{:d}, {:02d}:{:02d}:{:02d}".format(years, months, days, hours, minutes, seconds)
+		else:
+			return "Year {:d}, {:d}-{:d}, {:02d}:{:02d}:{:02d}".format(years, months, days, hours, minutes, seconds)
+	
+
 	def ReadWithCSPulse(self):
 		return False
 
 	def WriteWithCSPulse(self):
 		return True
 
+	def GetMaxROMSize(self):
+		return 512*1024
+
 class DMG_Unlicensed_256M(DMG_MBC5):
 	def GetName(self):
 		return "256M Multi Cart"
 	
-	def SelectBankROM(self, index):
-		dprint(self.GetName(), "|", index)
+	def HasFlashBanks(self):
+		return True
 
+	def SelectBankFlash(self, index):
 		flash_bank = math.floor(index / 512)
-		index = index % 512
+		dprint(self.GetName(), "|SelectBankFlash()|", index, "->", flash_bank)
 
-		if index == 0:
-			self.CART_POWERCYCLE_FNCPTR()
-			commands = [
-				[ 0x7000, 0x00 ],
-				[ 0x7001, 0x00 ],
-				[ 0x7002, 0x90 + flash_bank ]
-			]
-			self.CartWrite(commands, delay=0.1)
+		self.CART_POWERCYCLE_FNCPTR()
 		
 		commands = [
-			[ 0x2100, index & 0xFF ],
+			[ 0x7000, 0x00 ],
+			[ 0x7001, 0x00 ],
+			[ 0x7002, 0x90 + flash_bank ]
+		]
+		self.CURRENT_FLASH_BANK = flash_bank
+		self.CartWrite(commands, delay=0.1)
+
+	def SelectBankROM(self, index):
+		dprint(self.GetName(), "|SelectBankROM()|", index)
+
+		if (index % 512 == 0) or (self.CURRENT_FLASH_BANK != math.floor(index / 512)):
+			self.SelectBankFlash(index)
+		self.CURRENT_ROM_BANK = index
+		index = index % 512
+
+		commands = [
 			[ 0x3000, ((index >> 8) & 0xFF) ],
+			[ 0x2100, index & 0xFF ],
 		]
 		
 		start_address = 0 if index == 0 else 0x4000
@@ -1192,6 +1347,9 @@ class DMG_Unlicensed_256M(DMG_MBC5):
 
 		return (start_address, self.RAM_BANK_SIZE)
 
+	def GetMaxROMSize(self):
+		return 32*1024*1024
+
 class DMG_Unlicensed_WisdomTree(DMG_MBC):
 	def GetName(self):
 		return "Wisdom Tree"
@@ -1209,6 +1367,9 @@ class DMG_Unlicensed_WisdomTree(DMG_MBC):
 		self.CartWrite(commands)
 		return (0, 0x8000)
 
+	def GetMaxROMSize(self):
+		return 2*1024*1024
+
 class DMG_Unlicensed_XploderGB(DMG_MBC):
 	def GetName(self):
 		return "Xploder GB"
@@ -1234,6 +1395,9 @@ class DMG_Unlicensed_XploderGB(DMG_MBC):
 			self.CartRead(0x0102, 1)
 		return self.SelectBankROM(index + 8)
 
+	def GetMaxROMSize(self):
+		return 128*1024
+
 class DMG_Unlicensed_Sachen(DMG_MBC):
 	def GetName(self):
 		return "Sachen"
@@ -1247,6 +1411,9 @@ class DMG_Unlicensed_Sachen(DMG_MBC):
 		start_address = 0x4000
 		return (start_address, self.ROM_BANK_SIZE)
 
+	def GetMaxROMSize(self):
+		return 2*1024*1024
+
 class DMG_Unlicensed_DatelOrbitV2(DMG_MBC):
 	def GetName(self):
 		return "Datel Orbit V2"
@@ -1266,12 +1433,41 @@ class DMG_Unlicensed_DatelOrbitV2(DMG_MBC):
 		start_address = 0x4000
 		return (start_address, self.ROM_BANK_SIZE)
 
+	def GetMaxROMSize(self):
+		return 128*1024
+
+# class DMG_Unlicensed_DatelMegaMem(DMG_MBC):
+# 	def GetName(self):
+# 		return "Datel MegaMem"
+	
+# 	def __init__(self, args=None, cart_write_fncptr=None, cart_read_fncptr=None, cart_powercycle_fncptr=None, clk_toggle_fncptr=None):
+# 		if args is None: args = {}
+# 		super().__init__(args=args, cart_write_fncptr=cart_write_fncptr, cart_read_fncptr=cart_read_fncptr, cart_powercycle_fncptr=cart_powercycle_fncptr, clk_toggle_fncptr=None)
+# 		self.ROM_BANK_SIZE = 0x4000
+# 		self.RAM_BANK_SIZE = 0x4000
+
+# 	def SelectBankROM(self, index):
+# 		dprint(self.GetName(), "|", index)
+# 		return (0, self.ROM_BANK_SIZE)
+
+# 	def SelectBankRAM(self, index):
+# 		dprint(self.GetName(), "|", index)
+# 		self.CartWrite([[ 0x2000, index & 0x20 ]])
+# 		return (0x4000, self.RAM_BANK_SIZE)
+
+# 	def GetROMBanks(self, rom_size):
+# 		return 1
+	
+# 	def GetMaxROMSize(self):
+# 		return 16*1024
+
 
 class AGB_GPIO:
 	CART_WRITE_FNCPTR = None
 	CART_READ_FNCPTR = None
 	CLK_TOGGLE_FNCPTR = None
 	RTC = False
+	RTC_BUFFER = None
 
 	# Addresses
 	GPIO_REG_DAT = 0xC4 # Data
@@ -1443,8 +1639,7 @@ class AGB_GPIO:
 		buffer.append(self.RTCReadStatus()) # 24h mode = 0x40, reset flag = 0x80
 		buffer.extend(struct.pack("<Q", ts))
 
-		dstr = ' '.join(format(x, '02X') for x in buffer)
-		dprint("[{:02X}] {:s}".format(int(len(dstr)/3) + 1, dstr))
+		dprint(' '.join(format(x, '02X') for x in buffer))
 		
 		# Digits are BCD (Binary Coded Decimal)
 		#[07] 00 01 27 05 06 30 20
@@ -1452,9 +1647,11 @@ class AGB_GPIO:
 		# "27 days, 06:51:55"
 		#[07] 00 01 27 05 06 52 18
 		#     YY MM DD WW HH MM SS
+		self.RTC_BUFFER = buffer
 		return buffer
 	
 	def WriteRTC(self, buffer, advance=False):
+		rtc_status = None
 		if advance:
 			try:
 				dt_now = datetime.datetime.fromtimestamp(time.time())
@@ -1482,7 +1679,7 @@ class AGB_GPIO:
 					timestamp_now = int(time.time())
 					if timestamp_then < timestamp_now:
 						dt_then = datetime.datetime.fromtimestamp(timestamp_then)
-						dt_buffer = datetime.datetime.strptime("{:04d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}".format(years + 2000, months % 12, days % 31, hours % 60, minutes % 60, seconds % 60), "%Y-%m-%d %H:%M:%S")
+						dt_buffer = datetime.datetime.strptime("{:04d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}".format(years + 2000, months % 13, days % 32, hours % 60, minutes % 60, seconds % 60), "%Y-%m-%d %H:%M:%S")
 						rd = relativedelta(dt_now, dt_then)
 						dt_new = dt_buffer + rd
 						years = dt_new.year - 2000
@@ -1528,4 +1725,32 @@ class AGB_GPIO:
 			[ self.GPIO_REG_RE, 0 ], # Disable RTC Mapping
 		])
 
-		self.RTCWriteStatus(rtc_status)
+		if rtc_status is not None:
+			self.RTCWriteStatus(rtc_status)
+
+	def GetRTCString(self):
+		has_rtc = self.HasRTC()
+		if has_rtc is not True:
+			if has_rtc is False:
+				return "Not available"
+			elif has_rtc == 1:
+				return "Not available / Battery dry"
+			elif has_rtc == 3:
+				return "Not available"
+		
+		if self.RTC_BUFFER is None:
+			self.ReadRTC()
+		rtc_buffer = self.RTC_BUFFER
+		#weekdays = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat']
+		rtc_y = (rtc_buffer[0] & 0x0F) + ((rtc_buffer[0] >> 4) * 10)
+		rtc_m = (rtc_buffer[1] & 0x0F) + ((rtc_buffer[1] >> 4) * 10)
+		rtc_d = (rtc_buffer[2] & 0x0F) + ((rtc_buffer[2] >> 4) * 10)
+		#rtc_w = (rtc_buffer[3] & 0x0F) + ((rtc_buffer[3] >> 4) * 10)
+		rtc_h = ((rtc_buffer[4] & 0x0F) + (((rtc_buffer[4] >> 4) & 0x7) * 10))
+		rtc_i = (rtc_buffer[5] & 0x0F) + ((rtc_buffer[5] >> 4) * 10)
+		rtc_s = (rtc_buffer[6] & 0x0F) + ((rtc_buffer[6] >> 4) * 10)
+
+		if rtc_y == 0 and rtc_m == 0 and rtc_d == 0 and rtc_h == 0 and rtc_i == 0 and rtc_s == 0:
+			raise Exception("Invalid RTC data")
+		else:
+			return "20{:02d}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}".format(rtc_y, rtc_m, rtc_d, rtc_h, rtc_i, rtc_s)
