@@ -7,25 +7,29 @@ from enum import Enum
 
 # Common constants
 APPNAME = "FlashGBX"
-VERSION_PEP440 = "3.19"
+VERSION_PEP440 = "3.32"
 VERSION = "v{:s}".format(VERSION_PEP440)
+VERSION_TIMESTAMP = 1690298775
 DEBUG = False
+DEBUG_LOG = []
+APP_PATH = ""
+CONFIG_PATH = ""
 
-AGB_Header_ROM_Sizes = [ "64 KB", "128 KB", "256 KB", "512 KB", "1 MB", "2 MB", "4 MB", "8 MB", "16 MB", "32 MB", "64 MB", "128 MB", "256 MB" ]
+AGB_Header_ROM_Sizes = [ "64 KiB", "128 KiB", "256 KiB", "512 KiB", "1 MiB", "2 MiB", "4 MiB", "8 MiB", "16 MiB", "32 MiB", "64 MiB", "128 MiB", "256 MiB" ]
 AGB_Header_ROM_Sizes_Map = [ 0x10000, 0x20000, 0x40000, 0x80000, 0x100000, 0x200000, 0x400000, 0x800000, 0x1000000, 0x2000000, 0x4000000, 0x8000000, 0x10000000 ]
-AGB_Header_Save_Types = [ "None", "4K EEPROM (512 Bytes)", "64K EEPROM (8 KB)", "256K SRAM/FRAM (32 KB)", "512K FLASH (64 KB)", "1M FLASH (128 KB)", "8M DACS (1008 KB)", "Unlicensed 512K SRAM (64 KB)", "Unlicensed 1M SRAM (128 KB)" ]
-AGB_Header_Save_Sizes = [ 0, 512, 8192, 32768, 65536, 131072, 1032192, 65536, 131072 ]
-AGB_Global_CRC32 = 0
-AGB_Flash_Save_Chips = { 0xBFD4:"SST 39VF512", 0x1F3D:"Atmel AT29LV512", 0xC21C:"Macronix MX29L512", 0x321B:"Panasonic MN63F805MNP", 0xC209:"Macronix MX29L010", 0x6213:"SANYO LE26FV10N1TS" }
-AGB_Flash_Save_Chips_Sizes = [ 0x10000, 0x10000, 0x10000, 0x10000, 0x20000, 0x20000 ]
+AGB_Header_Save_Types = [ "None", "4K EEPROM (512 Bytes)", "64K EEPROM (8 KiB)", "256K SRAM/FRAM (32 KiB)", "512K FLASH (64 KiB)", "1M FLASH (128 KiB)", "8M DACS (1 MiB)", "Unlicensed 512K SRAM (64 KiB)", "Unlicensed 1M SRAM (128 KiB)", "Unlicensed Batteryless SRAM" ]
+AGB_Header_Save_Sizes = [ 0, 512, 8192, 32768, 65536, 131072, 1048576, 65536, 131072, 0 ]
+AGB_Flash_Save_Chips = { 0xBFD4:"SST 39VF512", 0x1F3D:"Atmel AT29LV512", 0xC21C:"Macronix MX29L512", 0x321B:"Panasonic MN63F805MNP", 0xC209:"Macronix MX29L010", 0x6213:"SANYO LE26FV10N1TS", 0xBF5B:"Unlicensed SST SST49LF080A" }
+AGB_Flash_Save_Chips_Sizes = [ 0x10000, 0x10000, 0x10000, 0x10000, 0x20000, 0x20000, 0x20000 ]
 
-DMG_Header_Mapper = { 0x00:'None', 0x01:'MBC1', 0x02:'MBC1+SRAM', 0x03:'MBC1+SRAM+BATTERY', 0x06:'MBC2+SRAM+BATTERY', 0x10:'MBC3+RTC+SRAM+BATTERY', 0x13:'MBC3+SRAM+BATTERY', 0x19:'MBC5', 0x1A:'MBC5+SRAM', 0x1B:'MBC5+SRAM+BATTERY', 0x1C:'MBC5+RUMBLE', 0x1E:'MBC5+RUMBLE+SRAM+BATTERY', 0x20:'MBC6+SRAM+FLASH+BATTERY', 0x22:'MBC7+ACCELEROMETER+EEPROM', 0x101:'MBC1M', 0x103:'MBC1M+SRAM+BATTERY', 0x0B:'MMM01',  0x0D:'MMM01+SRAM+BATTERY', 0xFC:'GBD+SRAM+BATTERY', 0x105:'G-MMC1+SRAM+BATTERY', 0x104:'M161', 0xFF:'HuC-1+IR+SRAM+BATTERY', 0xFE:'HuC-3+RTC+SRAM+BATTERY', 0xFD:'TAMA5+RTC+EEPROM', 0x201:'Unlicensed 256M Mapper', 0x202:'Unlicensed Wisdom Tree Mapper', 0x203:'Unlicensed Xploder GB Mapper', 0x204:'Unlicensed Sachen Mapper', 0x205:'Unlicensed Datel Orbit V2 Mapper' }
-DMG_Header_ROM_Sizes = [ "32 KB", "64 KB", "128 KB", "256 KB", "512 KB", "1 MB", "2 MB", "4 MB", "8 MB", "16 MB", "32 MB" ]
+DMG_Header_Mapper = { 0x00:'None', 0x01:'MBC1', 0x02:'MBC1+SRAM', 0x03:'MBC1+SRAM+BATTERY', 0x06:'MBC2+SRAM+BATTERY', 0x0F:'MBC3+RTC+BATTERY', 0x10:'MBC3+RTC+SRAM+BATTERY', 0x110:'MBC30+RTC+SRAM+BATTERY', 0x12:'MBC3+SRAM', 0x13:'MBC3+SRAM+BATTERY', 0x19:'MBC5', 0x1A:'MBC5+SRAM', 0x1B:'MBC5+SRAM+BATTERY', 0x1C:'MBC5+RUMBLE', 0x1E:'MBC5+RUMBLE+SRAM+BATTERY', 0x20:'MBC6+SRAM+FLASH+BATTERY', 0x22:'MBC7+ACCELEROMETER+EEPROM', 0x101:'MBC1M', 0x103:'MBC1M+SRAM+BATTERY', 0x0B:'MMM01',  0x0D:'MMM01+SRAM+BATTERY', 0xFC:'MAC-GBD+SRAM+BATTERY', 0x105:'G-MMC1+SRAM+BATTERY', 0x104:'M161', 0xFF:'HuC-1+IR+SRAM+BATTERY', 0xFE:'HuC-3+RTC+SRAM+BATTERY', 0xFD:'TAMA5+RTC+EEPROM', 0x201:'Unlicensed 256M Mapper', 0x202:'Unlicensed Wisdom Tree Mapper', 0x203:'Unlicensed Xploder GB Mapper', 0x204:'Unlicensed Sachen Mapper', 0x205:'Unlicensed Datel Orbit V2 Mapper' }
+DMG_Mapper_Types = { "None":[ 0x00, 0x08, 0x09 ], "MBC1":[ 0x01, 0x02, 0x03 ], "MBC2":[ 0x05, 0x06 ], "MBC3":[ 0x0F, 0x10, 0x11, 0x12, 0x13 ], "MBC30":[ 0x110 ], "MBC5":[ 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E ], "MBC6":[ 0x20 ], "MBC7":[ 0x22 ], "MBC1M":[ 0x101, 0x103 ], "MMM01":[ 0x0B, 0x0D ], "MAC-GBD":[ 0xFC ], "G-MMC1":[ 0x105 ], "M161":[ 0x104 ], "HuC-1":[ 0xFF ], "HuC-3":[ 0xFE ], "TAMA5":[ 0xFD ], "Unlicensed 256M Multi Cart Mapper":[ 0x201 ], "Unlicensed Wisdom Tree Mapper":[ 0x202 ], "Unlicensed Xploder GB Mapper":[ 0x203 ], "Unlicensed Sachen Mapper":[ 0x204 ], "Unlicensed Datel Orbit V2 Mapper":[ 0x205 ] }
+DMG_Header_ROM_Sizes = [ "32 KiB", "64 KiB", "128 KiB", "256 KiB", "512 KiB", "1 MiB", "2 MiB", "4 MiB", "8 MiB", "16 MiB", "32 MiB" ]
 DMG_Header_ROM_Sizes_Map = [ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A ]
 DMG_Header_ROM_Sizes_Flasher_Map = [ 0x8000, 0x10000, 0x20000, 0x40000, 0x80000, 0x100000, 0x200000, 0x400000, 0x800000, 0x1000000, 0x2000000 ]
-DMG_Header_RAM_Sizes = [ "None", "4K SRAM (512 Bytes)", "16K SRAM (2 KB)", "64K SRAM (8 KB)", "256K SRAM (32 KB)", "512K SRAM (64 KB)", "1M SRAM (128 KB)", "MBC6 SRAM+FLASH (1.03 MB)", "MBC7 2K EEPROM (256 Bytes)", "MBC7 4K EEPROM (512 Bytes)", "TAMA5 EEPROM (32 Bytes)", "Unlicensed 4M SRAM (512 KB)", "Unlicensed 1M EEPROM (128 KB)" ]
-DMG_Header_RAM_Sizes_Map = [ 0x00, 0x01, 0x01, 0x02, 0x03, 0x05, 0x04, 0x104, 0x101, 0x102, 0x103, 0x201, 0x203 ]
-DMG_Header_RAM_Sizes_Flasher_Map = [ 0, 0x200, 0x800, 0x2000, 0x8000, 0x10000, 0x20000, 0x108000, 0x100, 0x200, 0x20, 0x80000, 0x20000 ] # RAM size in bytes
+DMG_Header_RAM_Sizes = [ "None", "4K SRAM (512 Bytes)", "16K SRAM (2 KiB)", "64K SRAM (8 KiB)", "256K SRAM (32 KiB)", "512K SRAM (64 KiB)", "1M SRAM (128 KiB)", "MBC6 SRAM+FLASH (1.03 MiB)", "MBC7 2K EEPROM (256 Bytes)", "MBC7 4K EEPROM (512 Bytes)", "TAMA5 EEPROM (32 Bytes)", "Unlicensed 4M SRAM (512 KiB)", "Unlicensed 1M EEPROM (128 KiB)" ]
+DMG_Header_RAM_Sizes_Map = [ 0x00, 0x100, 0x01, 0x02, 0x03, 0x05, 0x04, 0x104, 0x101, 0x102, 0x103, 0x201, 0x203, 0x204 ]
+DMG_Header_RAM_Sizes_Flasher_Map = [ 0, 0x200, 0x800, 0x2000, 0x8000, 0x10000, 0x20000, 0x108000, 0x100, 0x200, 0x20, 0x80000, 0x20000, 0x80000 ] # RAM size in bytes
 DMG_Header_SGB = { 0x00:'No support', 0x03:'Supported' }
 DMG_Header_CGB = { 0x00:'No support', 0x80:'Supported', 0xC0:'Required' }
 
@@ -48,11 +52,11 @@ class IniSettings():
 				if not os.path.isdir(os.path.dirname(path)):
 					os.makedirs(os.path.dirname(path))
 				if os.path.exists(path):
-					with open(path, "a+") as f: f.close()
+					with open(path, "a+", encoding="UTF-8") as f: f.close()
 				else:
-					with open(path, "w+") as f: f.close()
+					with open(path, "w+", encoding="UTF-8") as f: f.close()
 			except:
-				print("Error accessing the configuration directory or settings file.")
+				print("Can’t access the configuration directory or settings file.")
 				return
 			self.FILENAME = path
 			self.SETTINGS = configparser.RawConfigParser()
@@ -61,7 +65,7 @@ class IniSettings():
 				self.Reload()
 			except configparser.MissingSectionHeaderError:
 				print("Resetting invalid configuration file...")
-				os.unlink(path)
+				with open(path, "w+", encoding="UTF-8") as f: f.close()
 				path = ""
 		
 		if path == "":
@@ -75,7 +79,7 @@ class IniSettings():
 	def Reload(self):
 		if self.SETTINGS is None: return
 		if self.FILENAME is not False:
-			with open(self.FILENAME, "r", encoding="utf-8") as f:
+			with open(self.FILENAME, "r", encoding="UTF-8") as f:
 				self.SETTINGS.read_file(f)
 		if len(self.SETTINGS.sections()) == 0:
 			self.SETTINGS.add_section(self.MAIN_SECTION)
@@ -96,7 +100,7 @@ class IniSettings():
 		self.SETTINGS[self.MAIN_SECTION][key] = value
 		dprint("Updating settings:", key, "=", value)
 		if self.FILENAME is not False:
-			with open(self.FILENAME, "w", encoding="utf-8") as f:
+			with open(self.FILENAME, "w", encoding="UTF-8") as f:
 				self.SETTINGS.write(f)
 	
 	def clear(self): self.Clear()
@@ -104,7 +108,7 @@ class IniSettings():
 		if self.SETTINGS is None: return None
 		self.SETTINGS.clear()
 		if self.FILENAME is not False:
-			with open(self.FILENAME, "w", encoding="utf-8") as f:
+			with open(self.FILENAME, "w", encoding="UTF-8") as f:
 				self.SETTINGS.write(f)
 
 class Progress():
@@ -123,12 +127,16 @@ class Progress():
 			if args["action"] == "INITIALIZE":
 				self.PROGRESS["action"] = args["action"]
 				self.PROGRESS["method"] = args["method"]
+				if "flash_offset" in args:
+					self.PROGRESS["flash_offset"] = args["flash_offset"]
+				else:
+					self.PROGRESS["flash_offset"] = 0
 				if "size" in args:
-					self.PROGRESS["size"] = args["size"]
+					self.PROGRESS["size"] = args["size"] - self.PROGRESS["flash_offset"]
 				else:
 					self.PROGRESS["size"] = 0
 				if "pos" in args:
-					self.PROGRESS["pos"] = args["pos"]
+					self.PROGRESS["pos"] = args["pos"] - self.PROGRESS["flash_offset"]
 				else:
 					self.PROGRESS["pos"] = 0
 				if "time_start" in args:
@@ -147,7 +155,7 @@ class Progress():
 				self.UPDATER(args)
 				self.PROGRESS = {}
 			
-			elif args["action"] in ("ERASE", "SECTOR_ERASE", "UNLOCK", "UPDATE_RTC"):
+			elif args["action"] in ("ERASE", "SECTOR_ERASE", "UNLOCK", "UPDATE_RTC", "ERROR"):
 				if "time_start" in self.PROGRESS:
 					args["time_elapsed"] = now - self.PROGRESS["time_start"]
 				elif "time_start" in args:
@@ -160,7 +168,7 @@ class Progress():
 				return
 			
 			elif args["action"] == "UPDATE_POS":
-				self.PROGRESS["pos"] = args["pos"]
+				self.PROGRESS["pos"] = args["pos"] - self.PROGRESS["flash_offset"]
 				self.PROGRESS["action"] = "PROGRESS"
 				if "time_start" in self.PROGRESS:
 					self.PROGRESS["time_elapsed"] = now - self.PROGRESS["time_start"]
@@ -230,7 +238,7 @@ class Progress():
 		
 		finally:
 			self.MUTEX.release()
-
+	
 class TAMA5_CMD(Enum):
 	RAM_WRITE = 0x0
 	RAM_READ = 0x1
@@ -273,33 +281,30 @@ def isx2bin(buffer):
 			break
 	return data_output[:temp]
 
-def roundup(x):
-	# https://stackoverflow.com/questions/50405017/
-	d = 10 ** 2
-	if x < 0:
-		return math.floor(x * d) / d
-	else:
-		return math.ceil(x * d) / d
+def round2(num, decimals=2):
+	x = (pow(10, decimals))
+	return int(num * x) / x
 
-def formatFileSize(size, asInt=False, roundUp=False):
+def formatFileSize(size, asInt=False, nobr=True):
+	space = " " if nobr else " "
 	if size == 1:
-		return "{:d} Byte".format(size)
+		return "{:d}{:s}Byte".format(size, space)
 	elif size < 1024:
-		return "{:d} Bytes".format(size)
+		return "{:d}{:s}Bytes".format(size, space)
 	elif size < 1024 * 1024:
-		val = size/1024
-		if roundUp: val = roundup(val)
+		val = size / 1024
+		val = round2(val)
 		if asInt:
-			return "{:d} KB".format(int(val))
+			return "{:d}{:s}KiB".format(int(val), space)
 		else:
-			return "{:.1f} KB".format(val)
+			return "{:.1f}{:s}KiB".format(val, space)
 	else:
-		val = size/1024/1024
-		if roundUp: val = roundup(val)
+		val = size / 1024 / 1024
+		val = round2(val)
 		if asInt:
-			return "{:d} MB".format(int(val))
+			return "{:d}{:s}MiB".format(int(val), space)
 		else:
-			return "{:.2f} MB".format(val)
+			return "{:.2f}{:s}MiB".format(val, space)
 
 def formatProgressTimeShort(sec):
 	sec = sec % (24 * 3600)
@@ -309,11 +314,14 @@ def formatProgressTimeShort(sec):
 	sec %= 60
 	return "{:02d}:{:02d}:{:02d}".format(int(hr), int(min), int(sec))
 
-def formatProgressTime(sec):
+def formatProgressTime(sec, asFloat=False):
 	if int(sec) == 1:
 		return "{:d} second".format(int(sec))
 	elif sec < 60:
-		return "{:d} seconds".format(int(sec))
+		if sec < 1 and asFloat:
+			return "{:.2f} seconds".format(sec)
+		else:
+			return "{:d} seconds".format(int(sec))
 	elif int(sec) == 60:
 		return "1 minute"
 	else:
@@ -416,8 +424,6 @@ def ParseCFI(buffer):
 					info["tb_boot_sector"] = "{:s} (0x{:02X})".format(temp[buffer[pri_address + 0x1E]], buffer[pri_address + 0x1E])
 				except:
 					info["tb_boot_sector"] = "0x{:02X}".format(buffer[pri_address + 0x1E])
-		#elif "{:s}{:s}{:s}".format(chr(buffer[0x214]), chr(buffer[0x216]), chr(buffer[0x218])) == "PRI":
-		#	pass
 		
 		info["device_size"] = int(math.pow(2, buffer[0x4E]))
 		info["buffer_size"] = buffer[0x56] << 8 | buffer[0x54]
@@ -449,7 +455,29 @@ def ParseCFI(buffer):
 	
 	return info
 
+def ConvertMapperToMapperType(mapper_raw):
+	i = 0
+	for (k, v) in DMG_Mapper_Types.items():
+		if i == 0:
+			retval = (k, v, i)
+		if mapper_raw in v:
+			retval = (k, v, i)
+			break
+		i += 1
+	# (string, list of ids, index in DMG_Mapper_Types)
+	return retval
+
+def ConvertMapperTypeToMapper(mapper_type):
+	i = 0
+	for (_, v) in DMG_Mapper_Types.items():
+		if mapper_type == i:
+			return v[0]
+		i += 1
+	return 0
+
 def GetDumpReport(di, device):
+	header = di["header"]["unchanged"]
+	if "db" in di["header"]: header["db"] = di["header"]["db"]
 	if di["system"] == "DMG":
 		mode = "DMG"
 		di["system"] = "Game Boy"
@@ -459,7 +487,7 @@ def GetDumpReport(di, device):
 			di["rom_size"] = "{:,} bytes".format(di["rom_size"])
 		
 		if di["mapper_type"] in DMG_Header_Mapper:
-			di["mapper_type"] = "{:s}".format(DMG_Header_Mapper[di["mapper_type"]])
+			di["mapper_type"] = "{:s}".format(ConvertMapperToMapperType(di["mapper_type"])[0])
 		else:
 			di["mapper_type"] = "0x{:02X}".format(di["mapper_type"])
 	
@@ -476,7 +504,7 @@ def GetDumpReport(di, device):
 		di["file_name"] = ""
 	else:
 		di["file_name"] = os.path.split(di["file_name"])[1]
-	di["file_size"] = "{:s} ({:d} bytes)".format(formatFileSize(di["file_size"]), di["file_size"])
+	di["file_size"] = "{:s} ({:d} bytes)".format(formatFileSize(size=di["file_size"]), di["file_size"])
 	
 	s = "" \
 		"= FlashGBX Dump Report =\n" \
@@ -496,10 +524,12 @@ def GetDumpReport(di, device):
 		"* Hardware:        {hardware:s} – Firmware {firmware:s}\n" \
 		"* Software:        {software:s}\n" \
 		"* OS Platform:     {platform:s}\n" \
+		"* Baud Rate:       {baud_rate:d}\n" \
 		"* Dump Time:       {timestamp:s}\n" \
 		"* Time Elapsed:    %TIME_ELAPSED% (%TRANSFER_RATE%)\n" \
 		"* Transfer Buffer: {buffer_size:d} bytes\n" \
-	.format(hardware=device.GetFullName(), firmware=device.GetFirmwareVersion(), software="{:s} {:s}".format(APPNAME, VERSION), platform=platform.platform(), buffer_size=di["transfer_size"], timestamp=di["timestamp"])
+		"* Retries:         {retries:d}\n" \
+	.format(hardware=device.GetFullName(), firmware=device.GetFirmwareVersion(), software="{:s} {:s}".format(APPNAME, VERSION), platform=platform.platform(), buffer_size=di["transfer_size"], timestamp=di["timestamp"], baud_rate=device.GetBaudRate(), retries=device.GetReadErrors())
 	
 	if mode == "DMG":
 		s += "" \
@@ -517,43 +547,41 @@ def GetDumpReport(di, device):
 			"* Cartridge Type:  {cart_type:s}\n" \
 		.format(system=di["system"], rom_size=di["rom_size"], cart_type=di["cart_type"])
 
-	di["hdr_logo"] = "OK" if di["header"]["logo_correct"] else "Invalid"
-	di["header"]["game_title_raw"] = di["header"]["game_title_raw"].replace("\0", "␀")
+	di["hdr_logo"] = "OK" if header["logo_correct"] else "Invalid"
+	header["game_title_raw"] = header["game_title_raw"].replace("\0", "␀")
 	if mode == "DMG":
-		game_title = di["header"]["game_title_raw"]
+		game_title = header["game_title"]
 		game_code = ""
-		if di["header"]["cgb"] == 0xC0 or di["header"]["cgb"] == 0x80:
-			if len(di["header"]["game_title_raw"].rstrip("\x00")) == 15:
-				if game_title[-4:][0] in ("A", "B", "H", "K", "V") and game_title[-4:][3] in ("A", "B", "D", "E", "F", "I", "J", "K", "P", "S", "U", "X", "Y"):
-					game_code = "* Game Code:       {:s}\n".format(game_title[-4:])
-					game_title = game_title[:-4].rstrip("_")
+		if header["cgb"] in (0xC0, 0x80):
+			if "game_code" in header and len(header["game_code"]) > 0:
+				game_code = "* Game Code:       {:s}\n".format(header["game_code"])
 			di["hdr_target_platform"] = "Game Boy Color"
-		elif di["header"]["old_lic"] == 0x33 and di["header"]["sgb"] == 0x03:
+		elif header["old_lic"] == 0x33 and header["sgb"] == 0x03:
 			di["hdr_target_platform"] = "Super Game Boy"
 		else:
 			di["hdr_target_platform"] = "Game Boy"
 
-		if di["header"]["old_lic"] == 0x33 and di["header"]["sgb"] == 0x03:
+		if header["old_lic"] == 0x33 and header["sgb"] == 0x03:
 			di["hdr_sgb"] = "Supported"
 		else:
-			di["hdr_sgb"] = "No support" # (146h=0x{:02X}, 14Bh=0x{:02X})".format(di["header"]["sgb"], di["header"]["old_lic"])
-		if di["header"]["cgb"] in DMG_Header_CGB:
-			di["hdr_cgb"] = "{:s}".format(DMG_Header_CGB[di["header"]["cgb"]])
+			di["hdr_sgb"] = "No support" # (146h=0x{:02X}, 14Bh=0x{:02X})".format(header["sgb"], header["old_lic"])
+		if header["cgb"] in DMG_Header_CGB:
+			di["hdr_cgb"] = "{:s}".format(DMG_Header_CGB[header["cgb"]])
 		else:
-			di["hdr_cgb"] = "Unknown (0x{:02X})".format(di["header"]["cgb"])
+			di["hdr_cgb"] = "Unknown (0x{:02X})".format(header["cgb"])
 		
-		di["hdr_header_checksum"] = "OK (0x{:02X})".format(di["header"]["header_checksum"]) if di["header"]["header_checksum_correct"] else "Invalid (0x{:02X}≠0x{:02X})".format(di["header"]["header_checksum_calc"], di["header"]["header_checksum"])
-		di["header"]["rom_checksum_calc"] = device.INFO["rom_checksum_calc"]
-		di["header"]["rom_checksum_correct"] = di["header"]["rom_checksum_calc"] == di["header"]["rom_checksum"]
-		di["hdr_rom_checksum"] = "OK (0x{:04X})".format(di["header"]["rom_checksum"]) if di["header"]["rom_checksum_correct"] else "Invalid (0x{:04X}≠0x{:04X})".format(di["header"]["rom_checksum_calc"], di["header"]["rom_checksum"])
+		di["hdr_header_checksum"] = "OK (0x{:02X})".format(header["header_checksum"]) if header["header_checksum_correct"] else "Invalid (0x{:02X}≠0x{:02X})".format(header["header_checksum_calc"], header["header_checksum"])
+		header["rom_checksum_calc"] = device.INFO["rom_checksum_calc"]
+		header["rom_checksum_correct"] = header["rom_checksum_calc"] == header["rom_checksum"]
+		di["hdr_rom_checksum"] = "OK (0x{:04X})".format(header["rom_checksum"]) if header["rom_checksum_correct"] else "Invalid (0x{:04X}≠0x{:04X})".format(header["rom_checksum_calc"], header["rom_checksum"])
 		
-		di["hdr_rom_size"] = di["header"]["rom_size_raw"]
+		di["hdr_rom_size"] = header["rom_size_raw"]
 		if di["hdr_rom_size"] in DMG_Header_ROM_Sizes_Map:
 			di["hdr_rom_size"] = "{:s} (0x{:02X})".format(DMG_Header_ROM_Sizes[DMG_Header_ROM_Sizes_Map.index(di["hdr_rom_size"])], di["hdr_rom_size"])
 		else:
 			di["hdr_rom_size"] = "Unknown (0x{:02X})".format(di["hdr_rom_size"])
 		
-		di["hdr_save_type"] = di["header"]["ram_size_raw"]
+		di["hdr_save_type"] = header["ram_size_raw"]
 		if di["hdr_save_type"] == 0x00:
 			di["hdr_save_type"] = "No SRAM (0x{:02X})".format(di["hdr_save_type"])
 		elif di["hdr_save_type"] in DMG_Header_RAM_Sizes_Map:
@@ -561,7 +589,7 @@ def GetDumpReport(di, device):
 		else:
 			di["hdr_save_type"] = "Unknown (0x{:02X})".format(di["hdr_save_type"])
 		
-		di["hdr_mapper_type"] = di["header"]["mapper_raw"]
+		di["hdr_mapper_type"] = header["mapper_raw"]
 		if di["hdr_mapper_type"] in DMG_Header_Mapper:
 			di["hdr_mapper_type"] = "{:s} (0x{:02X})".format(DMG_Header_Mapper[di["hdr_mapper_type"]], di["hdr_mapper_type"])
 		else:
@@ -581,15 +609,99 @@ def GetDumpReport(di, device):
 			"* SRAM Size:       {hdr_save_type:s}\n" \
 			"* Mapper Type:     {hdr_mapper_type:s}\n" \
 			"* Target Platform: {hdr_target_platform:s}\n" \
-		.format(hdr_game_title=game_title, hdr_game_code=game_code, hdr_revision=str(di["header"]["version"]), hdr_sgb=di["hdr_sgb"], hdr_cgb=di["hdr_cgb"], hdr_logo=di["hdr_logo"], hdr_header_checksum=di["hdr_header_checksum"], hdr_rom_checksum=di["hdr_rom_checksum"], hdr_rom_size=di["hdr_rom_size"], hdr_save_type=di["hdr_save_type"], hdr_mapper_type=di["hdr_mapper_type"], hdr_target_platform=di["hdr_target_platform"])
+		.format(hdr_game_title=game_title, hdr_game_code=game_code, hdr_revision=str(header["version"]), hdr_sgb=di["hdr_sgb"], hdr_cgb=di["hdr_cgb"], hdr_logo=di["hdr_logo"], hdr_header_checksum=di["hdr_header_checksum"], hdr_rom_checksum=di["hdr_rom_checksum"], hdr_rom_size=di["hdr_rom_size"], hdr_save_type=di["hdr_save_type"], hdr_mapper_type=di["hdr_mapper_type"], hdr_target_platform=di["hdr_target_platform"])
 		if "gbmem" in di and di["gbmem"] is not None:
-			s += "" \
-				"* Map Parameters:  {gbmem:s}\n" \
-			.format(gbmem=''.join(format(x, '02X') for x in di["gbmem"][:0x18]))
-	
+			raw_data = ""
+			for i in range(0, 4):
+				raw_data += ''.join(format(x, '02X') for x in di["gbmem"][i*0x20:i*0x20+0x20]) + "\n                   "
+			raw_data = raw_data[:-20]
+			
+			if "gbmem_parsed" in di and di["gbmem_parsed"] is not None and len(di["gbmem_parsed"]) > 0:
+				if (isinstance(di["gbmem_parsed"], list)):
+					s += "" \
+						"\n== GB-Memory Data (Multi Menu) ==\n" \
+						"* Write Timestamp: {timestamp:s}\n" \
+						"* Write Kiosk ID:  {kiosk_id:s}\n" \
+						"* Number of Games: {num_games:d}\n" \
+						"* Write Counter:   {write_count:d}\n" \
+						"* Cartridge ID:    {cart_id:s}\n" \
+						"* Raw Map Data:    {raw_data:s}\n" \
+					.format(
+						timestamp=di["gbmem_parsed"][0]["timestamp"],
+						kiosk_id=di["gbmem_parsed"][0]["kiosk_id"],
+						cart_id=di["gbmem_parsed"][0]["cart_id"],
+						write_count=di["gbmem_parsed"][0]["write_count"],
+						num_games=di["gbmem_parsed"][0]["num_games"],
+						raw_data=raw_data
+					)
+					for i in range(1, len(di["gbmem_parsed"])):
+						if di["gbmem_parsed"][i]["menu_index"] == 0xFF: continue
+						if di["gbmem_parsed"][i]["header"]["logo_correct"] is False: continue
+						if i == 1:
+							s += "\n=== Menu ROM ===\n"
+						else:
+							s += "\n=== Game {:d} ===\n".format(i-1)
+						s += "" \
+							"* Game Code:       {game_code:s}\n" \
+							"* Game Title:      {title:s}\n" \
+							"* Write Timestamp: {timestamp:s}\n" \
+							"* Write Kiosk ID:  {kiosk_id:s}\n" \
+							"* Location:        {location:s}\n" \
+							"* ROM Size:        {size:s}\n" \
+						.format(
+							game_code=di["gbmem_parsed"][i]["game_code"],
+							title=di["gbmem_parsed"][i]["title"],
+							timestamp=di["gbmem_parsed"][i]["timestamp"],
+							kiosk_id=di["gbmem_parsed"][i]["kiosk_id"],
+							location="0x{:06X}–0x{:06X}".format(di["gbmem_parsed"][i]["rom_offset"], di["gbmem_parsed"][i]["rom_offset"]+di["gbmem_parsed"][i]["rom_size"]-1),
+							size="{:s} ({:d} bytes)".format(formatFileSize(size=di["gbmem_parsed"][i]["rom_size"]), di["gbmem_parsed"][i]["rom_size"]),
+						)
+						if "crc32" in di["gbmem_parsed"][i]: s += "* CRC32:           {:08x}\n".format(di["gbmem_parsed"][i]["crc32"])
+						if "md5" in di["gbmem_parsed"][i]: s += "* MD5:             {:s}\n".format(di["gbmem_parsed"][i]["md5"])
+						if "sha1" in di["gbmem_parsed"][i]: s += "* SHA-1:           {:s}\n".format(di["gbmem_parsed"][i]["sha1"])
+						if "sha256" in di["gbmem_parsed"][i]: s += "* SHA-256:         {:s}\n".format(di["gbmem_parsed"][i]["sha256"])
+						
+						if "db_entry" in di["gbmem_parsed"][i] and "crc32" in di["gbmem_parsed"][i] and di["gbmem_parsed"][i]["db_entry"]["rc"] == di["gbmem_parsed"][i]["crc32"]:
+							s += "* Database Match:  {:s} {:s}\n".format(di["gbmem_parsed"][i]["db_entry"]["gn"], di["gbmem_parsed"][i]["db_entry"]["ne"])
+				elif isinstance(di["gbmem_parsed"]["game_code"], str):
+					s += "" \
+						"\n== GB-Memory Data (Single Game) ==\n" \
+						"* Game Code:       {game_code:s}\n" \
+						"* Game Title:      {title:s}\n" \
+						"* Write Timestamp: {timestamp:s}\n" \
+						"* Write Kiosk ID:  {kiosk_id:s}\n" \
+						"* Write Counter:   {write_count:d}\n" \
+						"* Cartridge ID:    {cart_id:s}\n" \
+						"* Raw Map Data:    {raw_data:s}\n" \
+					.format(
+						game_code=di["gbmem_parsed"]["game_code"],
+						title=di["gbmem_parsed"]["title"],
+						timestamp=di["gbmem_parsed"]["timestamp"],
+						kiosk_id=di["gbmem_parsed"]["kiosk_id"],
+						cart_id=di["gbmem_parsed"]["cart_id"],
+						write_count=di["gbmem_parsed"]["write_count"],
+						raw_data=raw_data
+					)
+			else:
+				s += "" \
+					"* GB-Memory Data:  {:s}\n" \
+				.format(raw_data)
+		
+		if header["db"] is not None and header["db"]["rc"] == di["hash_crc32"]:
+			db = header["db"]
+			s += "\n== Database Match ==\n"
+			if "gn" in db and "ne" in db: s += "* Game Name:       {:s} {:s}\n".format(db["gn"], db["ne"])
+			elif "gn" in db: s += "* Game Name:       {:s}\n".format(db["gn"])
+			if "rg" in db: s += "* Region:          {:s}\n".format(db["rg"])
+			if "lg" in db: s += "* Language(s):     {:s}\n".format(db["lg"])
+			if "rv" in db: s += "* Revision:        {:s}\n".format(db["rv"])
+			if "gc" in db: s += "* Game Code:       {:s}\n".format(db["gc"])
+			if "rc" in db: s += "* ROM CRC32:       {:08x}\n".format(db["rc"])
+			if "rs" in db: s += "* ROM Size:        {:s}\n".format(formatFileSize(size=db["rs"], asInt=True))
+
 	if mode == "AGB":
-		di["header"]["game_code_raw"] = di["header"]["game_code_raw"].replace("\0", "␀")
-		di["hdr_header_checksum"] = "OK (0x{:02X})".format(di["header"]["header_checksum"]) if di["header"]["header_checksum_correct"] else "Invalid (0x{:02X}≠0x{:02X})".format(di["header"]["header_checksum_calc"], di["header"]["header_checksum"])
+		header["game_code_raw"] = header["game_code_raw"].replace("\0", "␀")
+		di["hdr_header_checksum"] = "OK (0x{:02X})".format(header["header_checksum"]) if header["header_checksum_correct"] else "Invalid (0x{:02X}≠0x{:02X})".format(header["header_checksum_calc"], header["header_checksum"])
 		if "agb_savelib" not in di:
 			di["agb_savelib"] = "None"
 		elif "SRAM_F_" in di["agb_savelib"]:
@@ -616,34 +728,57 @@ def GetDumpReport(di, device):
 			"* Nintendo Logo:   {hdr_logo:s}\n" \
 			"* Header Checksum: {hdr_header_checksum:s}\n" \
 			"* Save Type:       {agb_savetype}\n" \
-		.format(hdr_game_title=di["header"]["game_title_raw"], hdr_game_code=di["header"]["game_code_raw"], hdr_revision=str(di["header"]["version"]), hdr_logo=di["hdr_logo"], hdr_header_checksum=di["hdr_header_checksum"], agb_savetype=di["agb_savelib"])
+		.format(hdr_game_title=header["game_title_raw"], hdr_game_code=header["game_code_raw"], hdr_revision=str(header["version"]), hdr_logo=di["hdr_logo"], hdr_header_checksum=di["hdr_header_checksum"], agb_savetype=di["agb_savelib"])
 		if "agb_save_flash_id" in di and di["agb_save_flash_id"] is not None:
 			s += "" \
 				"* Save Flash Chip: {agb_save_flash_chip_name:s} (0x{agb_save_flash_chip_id:04X})\n" \
 			.format(agb_save_flash_chip_name=di["agb_save_flash_id"][1], agb_save_flash_chip_id=di["agb_save_flash_id"][0])
-	
+
+		if "eeprom_data" in di:
+			s += "* EEPROM area:     {:s}…\n".format(''.join(format(x, '02X') for x in di["eeprom_data"]))
+
+		if header["db"] is not None and header["db"]["rc"] == di["hash_crc32"]:
+			db = header["db"]
+			s += "\n== Database Match ==\n"
+			if "gn" in db and "ne" in db: s += "* Game Name:       {:s} {:s}\n".format(db["gn"], db["ne"])
+			elif "gn" in db: s += "* Game Name:       {:s}\n".format(db["gn"])
+			if "rg" in db: s += "* Region:          {:s}\n".format(db["rg"])
+			if "lg" in db: s += "* Language(s):     {:s}\n".format(db["lg"])
+			if "rv" in db: s += "* Revision:        {:s}\n".format(db["rv"])
+			if "gc" in db: s += "* Game Code:       {:s}\n".format(db["gc"])
+			if "rc" in db: s += "* ROM CRC32:       {:08x}\n".format(db["rc"])
+			if "rs" in db: s += "* ROM Size:        {:s}\n".format(formatFileSize(size=db["rs"], asInt=True))
+			if "st" in db: s += "* Save Type:       {:s}\n".format(AGB_Header_Save_Types[db["st"]])
+			#if "ss" in db: s += "* Save Size:       {:s}\n".format(formatFileSize(size=db["ss"], asInt=True))
+
 	return s
 
-def GenerateFileName(mode, header, settings):
+def GenerateFileName(mode, header, settings=None):
+	fe_ni = True
+	if settings is not None:
+		fe_ni = settings.value(key="UseNoIntroFilenames", default="enabled").lower() == "enabled"
+	
 	path = "ROM"
 	if mode == "DMG":
 		path_title = header["game_title"]
 		path_code = ""
 		path_revision = str(header["version"])
 		path_extension = "bin"
-		path = "%TITLE%"
+		path = "%TITLE%-%REVISION%"
 		fe_sgb = "enabled"
 		if settings is not None:
 			path = settings.value(key="FileNameFormatDMG", default=path)
 			fe_sgb = settings.value(key="AutoFileExtensionSGB", default="enabled")
-		if header["cgb"] == 0xC0 or header["cgb"] == 0x80:
-			if len(header["game_title_raw"].rstrip("\x00")) == 15:
-				if path_title[-4:][0] in ("A", "B", "H", "K", "V") and path_title[-4:][3] in ("A", "B", "D", "E", "F", "I", "J", "K", "P", "S", "U", "X", "Y"):
-					path_code = path_title[-4:]
-					path_title = path_title[:-4].rstrip("_")
-					path = "%TITLE%_%CODE%-%REVISION%"
-					if settings is not None:
-						path = settings.value(key="FileNameFormatCGB", default=path)
+		
+		if len(header["game_code"]) > 0:
+			path_code = header["game_code"]
+			path = "%TITLE%_%CODE%-%REVISION%"
+			if settings is not None:
+				path = settings.value(key="FileNameFormatCGB", default=path)
+
+		if header["mapper_raw"] >= 0x200:
+			path = "%TITLE%"
+		if header["cgb"] in (0xC0, 0x80):
 			path_extension = "gbc"
 		elif header["old_lic"] == 0x33 and header["sgb"] == 0x03 and fe_sgb.lower() == "enabled":
 			path_extension = "sgb"
@@ -656,6 +791,12 @@ def GenerateFileName(mode, header, settings):
 			path = path.replace("%CODE%", path_code.strip())
 			path = path.replace("%REVISION%", path_revision)
 			path = re.sub(r"[<>:\"/\\|\?\*]", "_", path)
+			if get_mbc_name(header["mapper_raw"]) == "G-MMC1":
+				if "gbmem_parsed" in header:
+					if (isinstance(header["gbmem_parsed"], list)):
+						path += "_{:s}".format(header["gbmem_parsed"][0]["cart_id"])
+					else:
+						path += "_{:s}".format(header["gbmem_parsed"]["cart_id"])
 			path += ".{:s}".format(path_extension)
 	elif mode == "AGB":
 		path = "%TITLE%_%CODE%-%REVISION%"
@@ -664,16 +805,50 @@ def GenerateFileName(mode, header, settings):
 		path_title = header["game_title"]
 		path_code = header["game_code"]
 		path_revision = str(header["version"])
+		path_extension = "gba"
 		if (path_title == "" and path_code == ""):
-			path = "ROM.gba"
+			path = "ROM"
 		else:
 			path = path.replace("%TITLE%", path_title.strip())
 			path = path.replace("%CODE%", path_code.strip())
 			path = path.replace("%REVISION%", path_revision)
 			path = re.sub(r"[<>:\"/\\|\?\*]", "_", path)
-			path += ".gba"
+		path += "." + path_extension
+	
+	if fe_ni and header["db"] is not None:
+		if mode == "DMG" and get_mbc_name(header["mapper_raw"]) == "G-MMC1" and "gbmem_parsed" in header:
+			if (isinstance(header["gbmem_parsed"], list)):
+				path = "NP GB-Memory Cartridge ({:s}).{:s}".format(header["gbmem_parsed"][0]["cart_id"], path_extension)
+			else:
+				path = "NP GB-Memory Cartridge ({:s}).{:s}".format(header["gbmem_parsed"]["cart_id"], path_extension)
+		else:
+			path = "{:s} {:s}.{:s}".format(header["db"]["gn"], header["db"]["ne"], path_extension)
+	
 	return path
 
+def compare_mbc(a, b):
+	for v in DMG_Mapper_Types.values():
+		if a in v and b in v: return True
+	return False
+
+def get_mbc_name(id):
+	for (k, v) in DMG_Mapper_Types.items():
+		if id in v: return k
+	return "Unknown mapper type 0x{:02X}".format(id)
+
+def save_size_includes_rtc(mode, mbc, save_size, save_type):
+	rtc_size = 0x10
+	if mode == "DMG":
+		save_type = DMG_Header_RAM_Sizes_Map.index(save_type)
+		if get_mbc_name(mbc) in ("MBC3", "MBC30"): rtc_size = 0x30
+		elif get_mbc_name(mbc) == "HuC-3": rtc_size = 0x0C
+		elif get_mbc_name(mbc) == "TAMA5": rtc_size = 0x28
+		return (((DMG_Header_RAM_Sizes_Flasher_Map[save_type] + rtc_size) % save_size) == 0)
+	elif mode == "AGB":
+		rtc_size = 0x10
+		return (((AGB_Header_Save_Sizes[save_type] + rtc_size) % save_size) == 0)
+	return False
+
 def validate_datetime_format(string, format):
 	try:
 		if string != datetime.datetime.strptime(string, format).strftime(format):
@@ -692,7 +867,12 @@ def find_size(data, max_size, min_size=0x20):
 	return offset
 
 def dprint(*args, **kwargs):
+	global DEBUG_LOG
+	stack = traceback.extract_stack()
+	stack = stack[len(stack)-2]
+	msg = "[{:s}] [{:s}:{:d}] {:s}(): {:s}".format(datetime.datetime.now().astimezone().replace(microsecond=0).isoformat(), stack.filename[stack.filename.replace("\\", "/").rindex("/")+1:], stack.lineno, stack.name, " ".join(map(str, args)), **kwargs)
+	DEBUG_LOG.append(msg)
+	DEBUG_LOG = DEBUG_LOG[-32768:]
 	if DEBUG:
-		stack = traceback.extract_stack()
-		stack = stack[len(stack)-2]
-		print("{:s}[{:s}] [{:s}:{:d}] {:s}(): {:s}".format(ANSI.CLEAR_LINE, datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"), stack.filename[stack.filename.replace("\\", "/").rindex("/")+1:], stack.lineno, stack.name, " ".join(map(str, args)), **kwargs))
+		msg = "{:s}{:s}".format(ANSI.CLEAR_LINE, msg)
+		print(msg)
